 /* 	disasm80.c	(c) by bill beech, 2011
	dis1802.c   Feb 1 to 3 2017 HRJ

DESCRIPTION
	Disassemble an executable 8080/8085 - er, 1802 - file.

MODIFICATION HISTORY
	11 Jan 11 -- original
        12 Jan 12 -- Fix bug with ''' construct.
        26 Jan 17 -- Fixed address table dump.
        27 Jan 17 -- V 1.2D Removed USED from addsym, fixed HEX/ASCII dump,
                     fixed O command for non 0-based files,
                     fixed to display command invocation.
        30 Jan 17 HRJ - V C1.2D 1802 codes
        7-14 Feb 17 HRJ - add more features
	8 Mar 17 HRJ - skp 38H and long skp 0C8H is one byte!
       22 Oct 18 HRJ - enhance CALL with DB
	25 nov 18 HRJ - option for "command" to be more than one char; minor fixes
     10 Jun 2020 HRJ - correct mnemonic for 7DH is "SDBI"
                     - replaced ??? with INP 0 for 68H
                     - other mnemonic corrects
				- added date, new version to startup string
      11 Feb 2025 HRJ - SEQ and REQ opcodes backwards!
      11 Feb 2025 HRJ - if branch addr byte at XX00H, branch is to following page!
	  09 Mar 2025 HRJ - SEP vs CALL changes
	 18 Mar 2025 HRJ - misspelled SCRT, must be "RETN"
*/

#include	<stdio.h>
#include	<string.h>
#include	<ctype.h>
#include	<stdlib.h>

#define	version	"1.32"
#define	adate "Mar 2025"

#define HI(x)   (x >> 8)
#define LO(x)   (x & 0xff)

//type definitions
typedef struct decode   {
    struct  decode  *next;
    char    type;
    unsigned short start;
    unsigned short end;
}   DECODE;

typedef struct sym   {
    struct  sym  *next;
    unsigned int count;
    char used;
    char type;
    unsigned short addr;
}   SYM;

#define USED    1
#define UNUSED  0

//prototypes
void disasm(void);
void pass1(void);
void addsym(char type, unsigned short addr);
void pass2(void);
SYM *getsym(unsigned short addr);
void dumpsym(void);
void dumptab(void);
void dumphex(void);
void command(void);
void setcode(unsigned char type, unsigned short addr);

//globals
FILE    *fi, *fc, *fo;
long    fsi;
unsigned int sof = 0, cnt = 1, eof = 0, mem = 0, rom = 0xf800;
// HRJ sof is "ORG" of binary file, from CMD (default 0)
DECODE	*dec_head = NULL;
SYM	*sym_head = NULL;
int call_reg = -1, ret_reg = -1;


char    *opc[256]   =
   {
/*    1           2           3           4                */
    "IDL\t",    "LDN\tR1",  "LDN\tR2",   "LDN\tR3",   /* 00   */
    "LDN\tR4",   "LDN\tR5",   "LDN\tR6",  "LDN\tR7",
    "LDN\tR8",    "LDN\tR9",   "LDN\tRA",   "LDN\tRB",
    "LDN\tRC",   "LDN\tRD",   "LDN\tRE",  "LDN\tRF",

    "INC\tR0",    "INC\tR1",  "INC\tR2",  "INC\tR3",   /* 10   */
    "INC\tR4",   "INC\tR5",   "INC\tR6",  "INC\tR7",
    "INC\tR8",    "INC\tR9",   "INC\tRA",  "INC\tRB",
    "INC\tRC",   "INC\tRD",   "INC\tRE",  "INC\tRF",

    "DEC\tR0",    "DEC\tR1",  "DEC\tR2",   "DEC\tR3",   /* 20   */
    "DEC\tR4",   "DEC\tR5",   "DEC\tR6",  "DEC\tR7",
    "DEC\tR8",    "DEC\tR9",   "DEC\tRA",   "DEC\tRB",
    "DEC\tRC",   "DEC\tRD",   "DEC\tRE",  "DEC\tRF",

    "BR\t",    "BQ\t", "BZ\t",    "BDF\t",  /* 30   */
    "B1\t",    "B2\t",   "B3\t",  "B4\t",
    "SKP\t",   "BNQ\t",  "BNZ\t",    "BNF\t",
    "BN1\t",   "BN2\t",   "BN3\t",  "BN4\t",
/*    0           1           2           3                 */
    "LDA\tR0", "LDA\tR1", "LDA\tR2", "LDA\tR3", /* 40   */
    "LDA\tR4", "LDA\tR5", "LDA\tR6", "LDA\tR7",
    "LDA\tR8", "LDA\tR9", "LDA\tRA", "LDA\tRB",
    "LDA\tRC", "LDA\tRD", "LDA\tRE", "LDA\tRF",

    "STR\tR0", "STR\tR1", "STR\tR2", "STR\tR3", /* 50   */
    "STR\tR4", "STR\tR5", "STR\tR6", "STR\tR7",
    "STR\tR8", "STR\tR9", "STR\tRA", "STR\tRB",
    "STR\tRC", "STR\tRD", "STR\tRE", "STR\tRF",

    "IRX\t", "OUT\t1", "OUT\t2", "OUT\t3", /* 60   */
    "OUT\t4", "OUT\t5", "OUT\t6", "OUT\t7",
    "INP\t0", "INP\t1", "INP\t2", "INP\t3",
    "INP\t4", "INP\t5", "INP\t6", "INP\t7",

    "RET\t", "DIS\t", "LDXA\t", "STXD\t", /* 70   */
    "ADC\t", "SDB\t", "SHRC\t",    "SMB\t",
    "SAV\t", "MARK\t", "REQ\t", "SEQ\t",
    "ADCI\t", "SDBI\t", "SHLC\t", "SMBI\t",
/*    0           1           2           3                 */
    "GLO\tR0",   "GLO\tR1",   "GLO\tR2",   "GLO\tR3",   /* 80   */
    "GLO\tR4",   "GLO\tR5",   "GLO\tR6",   "GLO\tR7",
    "GLO\tR8",   "GLO\tR9",   "GLO\tRA",   "GLO\tRB",
    "GLO\tRC",   "GLO\tRD",   "GLO\tRE",   "GLO\tRF",

    "GHI\tR0",   "GHI\tR1",   "GHI\tR2",   "GHI\tR3",   /* 90   */
    "GHI\tR4",   "GHI\tR5",   "GHI\tR6",   "GHI\tR7",
    "GHI\tR8",   "GHI\tR9",   "GHI\tRA",   "GHI\tRB",
    "GHI\tRC",   "GHI\tRD",   "GHI\tRE",   "GHI\tRF",

    "PLO\tR0",   "PLO\tR1",   "PLO\tR2",   "PLO\tR3",   /* A0   */
    "PLO\tR4",   "PLO\tR5",   "PLO\tR6",   "PLO\tR7",
    "PLO\tR8",   "PLO\tR9",   "PLO\tRA",   "PLO\tRB",
    "PLO\tRC",   "PLO\tRD",   "PLO\tRE",   "PLO\tRF",

    "PHI\tR0",   "PHI\tR1",   "PHI\tR2",   "PHI\tR3",   /* B0   */
    "PHI\tR4",   "PHI\tR5",   "PHI\tR6",   "PHI\tR7",
    "PHI\tR8",   "PHI\tR9",   "PHI\tRA",   "PHI\tRB",
    "PHI\tRC",   "PHI\tRD",   "PHI\tRE",   "PHI\tRF",
/*    0           1           2           3                 */
    "LBR\t",    "LBQ\t",   "LBZ\t",    "LBDF\t",    /* C0   */
    "NOP\t",    "LSNQ\t",  "LSNZ\t",    "LSNF\t",
    "LSKP\t",   "LBNQ\t",    "LBNZ\t",     "LBNF\t",
    "LSIE\t",   "LSQ\t",   "LSZ\t",    "LSDF\t",

    "SEP\tR0",    "SEP\tR1",   "SEP\tR2",    "SEP\tR3",    /* D0   */
    "SEP\tR4",    "SEP\tR5",  "SEP\tR6",    "SEP\tR7",
    "SEP\tR8",    "SEP\tR9",    "SEP\tRA",     "SEP\tRB",
    "SEP\tRC",    "SEP\tRD",    "SEP\tRE",    "SEP\tRF",

    "SEX\tR0",    "SEX\tR1",   "SEX\tR2",    "SEX\tR3",   /* E0   */
    "SEX\tR4",    "SEX\tR5",  "SEX\tR6",    "SEX\tR7",
    "SEX\tR8",    "SEX\tR9",   "SEX\tRA",    "SEX\tRB",
    "SEX\tRC",    "SEX\tRD",    "SEX\tRE",    "SEX\tRF",

    "LDX\t",     "OR\t",   "AND\t",     "XOR\t",     /* F0   */
    "ADD\t",     "SD\t",   "SHR\t",    "SM\t",
    "LDI\t",     "ORI\t",   "ANI\t",     "XRI\t",
    "ADI\t",     "SDI\t",    "SHL\t",    "SMI\t"
   };

char    opc_size[]  =
   {
/*  0   1   2   3   4   5   6   7   8   9   A   B   C   D   E   F           */
    1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  /* 00   */
    1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  /* 10   */
    1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  /* 20   */
    2,  2,  2,  2,  2,  2,  2,  2,  1,  2,  2,  2,  2,  2,  2,  2,  /* 30   */

    1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  /* 40   */
    1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  /* 50   */
    1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  /* 60   */
    1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  2,  2,  1,  2,  /* 70   */

    1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  /* 80   */
    1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  /* 90   */
    1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  /* A0   */
    1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  /* B0   */

    3,  3,  3,  3,  1,  1,  1,  1,  1,  3,  3,  3,  1,  1,  1,  1,  /* C0   */
    1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  /* D0   */
    1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  /* E0   */
    1,  1,  1,  1,  1,  1,  1,  1,  2,  2,  2,  2,  2,  2,  1,  2,  /* F0   */
   };

int main(int argc, char **argv)
{
    char fname[128];

    printf("%s 1802 Version %s date %s starting...\n", argv[0], version, adate); //sign on
    if (argc != 2) {		//command error
	printf("No file name specified\n");
	exit(1);
    }
    strcpy(fname,argv[1]);	//get source file
    strcat(fname,".bin");
    if ((fi = fopen(fname,"rb+")) == NULL) { //can open for read?
        printf("Could not open source file %s\n", fname); //no
        exit(1);
    }
    fseek(fi, 0, SEEK_END);	//get actual file size
    fsi = ftell(fi);
    eof = (unsigned int) fsi - 1;
    mem = eof + 1;
    fseek(fi, 0, SEEK_SET);	//point to begining
    strcpy(fname,argv[1]);	//create output file
    strcat(fname,".asm");
    if ((fo = fopen(fname,"w")) == NULL) { //can open for write?
        printf("Could not create output file %s\n", fname); //no
        exit(1);
    }
    strcpy(fname,argv[1]);	//open configuration file
    strcat(fname,".cfg");
    if ((fc = fopen(fname,"r")) == NULL) { //can open for read?
        printf("Could not open configuration file %s\n", fname); //no
        exit(1);
    }
    fprintf(fo,"; Disassembly of %s.bin\n", argv[1]);
    fprintf(fo,"; File is %d[%04X] bytes\n;\n", eof-1, eof-1);
    command();                  //process CFG file
    dumptab();                  //dump config table
    disasm();                   //do the disassembly
    fclose(fc);                 //close configuration file
    fclose(fo);                 //close output file
    fclose(fi);			//close source file
    printf("%s finished!\n", argv[0]); //sign off
    return 0;
}

/* this is a two pass disassembler.
    pass 1 - builds a table of label addresses used in operands
    pass 2 - disassembles code, tables, and data and replaces hexadecimal
    addresses with labels constructed from the address in address table.
    places label in the emitted source code where they belong.
    */

void disasm(void)
{
    pass1();
    pass2();
}

void pass1(void)
{
    DECODE *dptr;
    SYM *sptr;
    unsigned short offset = 0, temp; //HRJ offset is current location in binary file?
    unsigned char opcode;
    char type;
    int i, mflag = 0;

    printf("Pass 1\n");
    dptr = dec_head;			//start at top of decode table
    fseek(fi, 0, SEEK_SET);		//point to begining of source
    while (dptr) {			//while more entries
        offset = dptr->start;		//offset into 0-based file
        fseek(fi, offset - sof, SEEK_SET); //point to start of segment in file
        if (dptr->type == 'C') {        //code segment
            while (offset < dptr->end) { //until end of code segment
                opcode = fgetc(fi);     //get opcode
                switch(opc_size[opcode]) {
                    case 1:             //1-byte opcode
                        //if (call_reg >= 0) // process D0-DF opcodes as calls
                        //if (ret_reg >= 0)  // process D0-DF opcodes as returns
 			            if ((opcode >= 0xD0) && (opcode <= 0xDF)) { //SEP instruction
                             if ( (call_reg > 0) && (opcode == (0xD0+call_reg)) ) {
					            temp = (fgetc(fi)<< 8) + fgetc(fi); //get hi-lo address
					            addsym ('S',temp); //Subroutine addr
					            offset+=3; //
				             } else offset++;
			            }
	                    else offset++;       //step binary file pointer
                        break;
                    case 2:             //2-byte opcode
					    //offset+sof is absolute addr
                        temp = fgetc(fi); //HRJ get branch (low-byte) address
                        if ((opcode >= 0x30) && (opcode <= 0x3F)) { //HRJ branch
                            if (opcode != 0x38) { // not skip
				                temp = (offset & 0xFF00) + temp;
				                if ((offset & 0x00FF) == 0xFF) temp = temp + 0x100; /* jumps page*/
				                     //HRJ find memory page address, add branch addr
                                addsym('R', temp);
			                }                            }
                        //else addsym('A', temp); //anything else, not with 1802!
			            offset+=2;      //step binary file pointer
                        break;
                    case 3:             //3-byte opcode
                        offset+=3;      //step binary file pointer
                        temp = (fgetc(fi)<< 8) + fgetc(fi); //get hi-lo address
                        if  ( ((opcode >= 0xC0) && (opcode <= 0xC3)) ||
				        ((opcode >= 0xC8) && (opcode <= 0xCB))  ){ //HRJ long branch
                            addsym('J', temp); //JUMP
                        }
                        //else addsym('A', temp); //anything else, not with 1802!
                        break;
                }
            }
        } else if (dptr->type == 'M') {	//memory segment
            dptr->start = mem;
            dptr->end = rom - 1;
            mflag = 1;
        } else if (dptr->type == 'T') {	//table segment
            while (offset < dptr->end) { //until end of table segment
                temp = (fgetc(fi) << 8) + fgetc(fi); //get hi-lo address
                addsym('T', temp);      //put it into table
                offset += 2;            //step binary file pointer
            }
        }
        dptr = dptr->next;              //get next config table entry
        fflush(fo);                     //write out what is done so far
    }
    fprintf(fo, "\n; sof=%04X eof=%04X mem=%04X rom=%04X mflag=%d\n", sof, eof, mem, rom, mflag);

    for (i=0; i<=0xFFFF; i++) {         //mark addresses external to module and
                                        //create an EQU for each
        if ((sptr = getsym(i)) != NULL) { //HRJ test for NULL, table entry found
			if (sptr->type == 'S'); // leave it type S
            else if (i < sof)
                sptr->type = 'L';       //address below module
            else if (i >= mem && i < rom)
                sptr->type = 'H';       //address above module
            else if (i >= rom && i <= 0xffff)
                sptr->type = 'R';       //address in ROM
            if (sptr->type == 'L' || sptr->type == 'R' || sptr->type == 'J' || sptr->type == 'S')
                fprintf(fo, "\n%c%04X\tEQU\t0%04XH\t\t;", sptr->type, i, i);
            if ((mflag == 0) && sptr->type == 'H')
                fprintf(fo, "\n%c%04X\tEQU\t0%04XH\t\t;", sptr->type, i, i);
        }
    }
    fprintf(fo, "\n\n");
    fprintf(fo, "; ASCII character equates\n\n");
    fprintf(fo, "TAB\tEQU\t009H\n");
    fprintf(fo, "LF\tEQU\t00AH\n");
    fprintf(fo, "CR\tEQU\t00DH\n");
	if (call_reg >= 0) fprintf(fo, ";CALL register is R%d\n", call_reg);
	if (ret_reg  >= 0) fprintf(fo, ";RET register is R%d\n", ret_reg);
    printf("End of Pass 1\n");
}

void pass2(void)
{
    DECODE *dptr;
    SYM *sptr;
    unsigned int byte;
    unsigned short offset = 0;
    unsigned short temp, kk;
    unsigned int i, k, x;
    unsigned char opcode, ch;
    int j=0;

    printf("Pass 2\n");
    dptr = dec_head;			//start at top of decode table
    fseek(fi, 0, SEEK_SET);		//rewind source
    while (dptr) {			//while more entries
	offset = dptr->start;	        //offset into file
	fseek(fi, offset - sof, SEEK_SET); //point to start of segment in file
        printf("     Decoding %c from %04X to %04X\n", dptr->type, offset, dptr->end);
        if (dptr->type == 'C') { //code segment
            fprintf(fo, "\n;*******************************\n");
	        fprintf(fo, "; Start code segment\n");
            //fprintf(fo, "\tCSEG\n");
            while (offset < dptr->end) { //until end of code segment
                if ((sptr = getsym(offset)) != NULL) { //HRJ test for NULL, address found, output label
                    if (sptr->type == 'C') {
                        fprintf(fo, "\n;*******************************\n");
                        fprintf(fo, "; Start Subroutine\n");
                    }
                    fprintf(fo, "%c%04X", sptr->type, offset);
                }
                opcode = fgetc(fi);     //get opcode, could test feof(fi)!=0
                // fprintf(fo, "\t%s", opc[opcode]); //print mnemonic
                switch (opc_size[opcode]) {
                    case 1:
				        if ((opcode >= 0xD0) && (opcode <= 0xDF)) { //SEP instruction
                        	if ( (call_reg > 0) && (opcode == (0xD0+call_reg)) ) { /* CALL */
							     temp = (fgetc(fi)<< 8) + fgetc(fi); //get hi-lo address
							      if ( (sptr = getsym(temp)) != NULL) ch = sptr->type; else ch='\0'; /* adr in sym tbl? */
                                 /* fprintf(fo, "\t\t;%04X-%02X CALL+addr DW\t%04XH\n", offset, opcode, temp); */
							     /* fseek(fi, -2L, SEEK_CUR); // move backwards two chars and disassemble */
								 fprintf(fo,"\tCALL\t%c%04X\t\t;%04X-%02X %02X %02X\n",ch,temp,offset,opcode,HI(temp),LO(temp)); /*CALL instr */
								 offset+=2;
                            }
                        	else if ( (ret_reg > 0) && (opcode == (0xD0+ret_reg)) ) { /* RET  */
								 /* fprintf(fo, "\t%s", opc[opcode]); //print mnemonic */
                                 /* fprintf(fo, "\t\t;%04X-%02X RET\n", offset, opcode); */
								 fprintf(fo,"\tRETN\t\t\t;%04X-%02X SEP R%01X\n",offset, opcode, ret_reg);
                            }
							else {
								 fprintf(fo, "\t%s", opc[opcode]); //print mnemonic
								 fprintf(fo, "\t\t;%04X-%02X\n", offset, opcode); // not CALL or RET
							}
						}
                        else {
							fprintf(fo, "\t%s", opc[opcode]); //print mnemonic
							fprintf(fo, "\t\t;%04X-%02X\n", offset, opcode); //not SEP
					    }
                        offset++;
                        break;
                    case 2:
						fprintf(fo, "\t%s", opc[opcode]); //print mnemonic
                        temp = fgetc(fi);
                        // HRJ if branch, print temp as absolute addr else
                        if ((opcode >= 0x30) && (opcode <= 0x3F)) { //HRJ branch
                            if (opcode != 0x38) { // not skip
				                //HRJ find memory page address, add branch
				                kk = temp + (offset & 0xFF00); //HRJ make abs addr
				                if ((offset & 0x00FF) == 0xFF) kk = kk + 0x100; /* jumps page*/
								if ( (sptr = getsym(kk)) != NULL) ch = sptr->type; else ch = 'R'; // address in symbol table
                                fprintf(fo, "%c%04X\t\t;%04X-%02X %02X\n", ch, kk, offset, opcode, temp);
			                } else {
				                kk = temp + (offset & 0xFF00);
				                fprintf(fo, " %04X\t\t;%04X-%02X %02X\n", kk, offset, opcode, temp);
			                }
                        }
                        else {
			                fprintf(fo, "0%02XH\t\t;%04X-%02X %02X", temp, offset, opcode, temp);
			                if (isalnum(temp)) fprintf(fo, " >%c<\n", temp); else fprintf(fo, "\n"); // print ASCII
                        }
			            offset+=2;
                        break;
                    case 3:
						fprintf(fo, "\t%s", opc[opcode]); //print mnemonic
                        temp = (fgetc(fi) << 8) + fgetc(fi);
                        if ( (sptr = getsym(temp)) != NULL) { //HRJ test for NULL,
                            // if (opcode == 0x31) {
                            //    fprintf(fo, "%c%04X\t;%04X-%02X %02X %02X\n",
                            //        sptr->type, temp, offset, opcode, HI(temp), LO(temp)); //HRJ
                            // } else {
                            fprintf(fo, "%c%04X\t\t;%04X-%02X %02X %02X\n",
                                    sptr->type, temp, offset, opcode, HI(temp), LO(temp)); // HRJ
						} else { //HRJ may not have decoded this in pass 1
							fprintf(fo,  "0%04XH\t;%04X-%02X %02X %02X\n",
									            temp, offset, opcode, HI(temp), LO(temp));
                        }
                        offset+=3;
                        break;
                } /* end case */
            } /* end while */
            fprintf(fo, "; End code segment\n");
        } else if (dptr->type == 'D') { //data segment
            fprintf(fo, "\n;*******************************\n");
            fprintf(fo, "; Start initialized data segment at %04X\n", offset);
            //  fprintf(fo, "\tDSEG\n");
            j = 9;                      //force new line, column 0
            while (offset < dptr->end) { //until end of data segment
                byte = fgetc(fi);       //get next byte
                if (byte == EOF) {      //uninitialized memory
                    offset++;
                } else {                //initialized memory
                    if ((sptr = getsym(offset)) != NULL) { //HRJ test for  NULL, address found, output label
						if ((j>0) && (j!=9)) fprintf(fo, "\t;%04X-%04X",offset-j, offset-1);  // display addr
                        fprintf(fo, "\n%c%04X\tDB\t", sptr->type, offset);
                        j = 0;          //new line, column 0
                    } else { //check end of data statement
                        if (j == 9) {
                            fprintf(fo, "\n\tDB\t");
                            j = 0;      //new line, column 0, at start
                        }
                    	else if (j > 7) {
                            fprintf(fo, "\t;%04X-%04X\n\tDB\t",offset-8, offset-1);
                            j = 0;      //display address, then new line, column 0
                        }
					}
                    if (j) {            // continuation value
                        if (isprint(byte) && byte != 0x27)
                            fprintf(fo, ", '%c'", byte); //print character
                        else {
                            if (byte == 0)
                                fprintf(fo, ", 0"); //print NUL
                            else if (byte == 0x9)
                                fprintf(fo, ", TAB"); //print TAB
                            else if (byte == 0xa)
                                fprintf(fo, ",  LF"); //print LF
                            else if (byte == 0xd)
                                fprintf(fo, ",  CR"); //print CR
                            else
                                fprintf(fo, ", 0%02XH", byte); //print everything else
                        }
                    }
                    else {              // initial value
                        if (isprint(byte) && byte != 0x27)
                            fprintf(fo, "'%c'", byte); //print character
                        else {
                            if (byte == 0)
                                fprintf(fo, "0"); //print NUL
                            else if (byte == 0x9)
                                fprintf(fo, "TAB"); //print TAB
                            else if (byte == 0xa)
                                fprintf(fo, " LF"); //print LF
                            else if (byte == 0xd)
                                fprintf(fo, " CR"); //print CR
                            else
                                fprintf(fo, "0%02XH", byte); //print data byte
                        }
                    }
                    offset++;
                    j++;                //next column
                }
            }
			if ((j>0) && (j!=9)) fprintf(fo, "\t;%04X-%04X",offset-j, offset-1);  // display addr
            fprintf(fo, "\n; End initialized data segment at %04X\n",offset-1);

        } else if (dptr->type == 'X') { //unit'ted  segment
            fprintf(fo, "\n;*******************************\n");
            fprintf(fo, "; Start uninitialized segment X\n");
			x = 0; // count space
            while (offset < dptr->end) { //until end of memory segment
                if ((sptr = getsym(offset)) != NULL) { // HRJ test for NULL, address found, output label
                    if (sptr->next != NULL) {
						fprintf(fo, "\tDS\t%d\t\t;%04X-%04X\n%c%04X",  x, offset-x, offset-1, sptr->type, offset);  //
						x = 1;
                    } else {
                        fprintf(fo, "%c%04X\tDS\t%d\n", sptr->type, offset,x); // end of symbols, dump space
						x = 0;
                    }
                } else {
					x++;  // count space up
				}
                offset++;
            }
			if (x > 0) fprintf (fo, "\tDS\t%d\t\t;%04X-%04X\n", x, offset-x, offset-1); // dump any space left
            fprintf(fo, "\n; End uninitialized  segment X\n");

        } else if (dptr->type == 'M') { //memory segment
            fprintf(fo, "\n;*******************************\n");
            fprintf(fo, "; Start uninitialized memory segment\n");
            while (offset < dptr->end) { //until end of memory segment
                if ((sptr = getsym(offset)) != NULL) { // HRJ test for NULL, address found, output label
                    if (sptr->next != NULL) {
                        fprintf(fo, "%c%04X\tDS\t%d\n", sptr->type, offset, sptr->next->addr - offset);
                    } else {
                        fprintf(fo, "%c%04X\tDS\t1\n", sptr->type, offset);
                    }
                }
                offset++;
            }
            fprintf(fo, "\n; End uninitialized memory segment\n");

        } else if (dptr->type == 'T') { //table segment
            fprintf(fo, "\n;*******************************\n");
	       fprintf(fo, "; Start table segment\n");
	       while (offset < dptr->end) { //until end of code segment
                if ((sptr = getsym(offset)) != NULL)//HRJ test for NULL, address found, output label
                     fprintf(fo, "%c%04X", sptr->type, offset);
                fprintf(fo, "\tDW\t");  //print mnemonic
                temp = (fgetc(fi) << 8) + fgetc(fi);
                if ((sptr = getsym(temp)) != NULL) { // HRJ test for NULL,
                    fprintf(fo, "%c%04X\t\t;%04X-%02X %02X *\n",
                        sptr->type, temp, offset, HI(temp), LO(temp)); //HRJ
                    offset += 2;
                }
            }
            fprintf(fo, "; End table segment\n");
	    fflush(fo);
	   } else if (dptr->type == 'S') //dump address table
                dumpsym();
        dptr = dptr->next;              //get next config table entry
        fflush(fo);                     //write out what is done so far
    }
	// while (feof(fi)==0) { //HRJ code or data left in binary file after last config entry
	// temp = fgetc(fi);
	// if (feof(fi)!=0) break;
	// fprintf(fo, "\tDB\t%02X\n", temp);
	// }

	if (feof(fi)==0) { //HRJ code or data left in file
		fprintf(fo,"data still available in object file\n");
	}
    fprintf(fo, "\tEND\n");
    printf("End of Pass 2\n");
}

/* add a new address (label) to address table */

void addsym(char type, unsigned short addr)
{
    SYM *nptr, *ptr, *pptr;

    if (ptr = getsym(addr)) {		//already in symbol table
        ptr->count++;			//step count
        return;
    }
    if ((nptr = (SYM *) malloc(sizeof(SYM))) == NULL) {
        printf("Unable to allocate memory for symbol table entry\n");
        exit(1);
    }
    nptr->next = NULL;			//initialize this entry
    nptr->used = 1;
    nptr->type = type;
    nptr->addr = addr;
    nptr->count = 1;
    if (sym_head == NULL)		//empty list, first entry
        sym_head = nptr;
    else {
        ptr = sym_head;			//start at top
        if (ptr->addr > addr) {		// if less then top of list, put on top
            nptr->next = ptr;
            sym_head = nptr;
        } else {
	    pptr = sym_head;
            while (ptr) {		//step through list
                if (ptr->addr > addr)
                    break;
                pptr = ptr;
                ptr = ptr->next;
            }
            nptr->next = ptr;		//link new one here
            pptr->next = nptr;
        }
    }
}

/* find address in address table */

SYM *getsym(unsigned short addr)
{
    SYM *ptr;

    ptr = sym_head;                     //start at top
    while (ptr) {
        if (ptr->addr == addr)
            return ptr;
        ptr = ptr->next;                //next entry
    }
    return NULL;
}

/* dump the symbol table */

void dumpsym(void)
{
    SYM *ptr;

    ptr = sym_head;                     //start at top
    fprintf(fo, "\n;*******************************\n");
    fprintf(fo, "; Start of Address Table\n");
    fprintf(fo, "; Type addr\tused?\tcount\n");
    while (ptr) {
        fprintf(fo, ";   %c  %04X\t%c\t%d\n",
            ptr->type, ptr->addr, ptr->used ? 'Y' : 'N', ptr->count);
        ptr = ptr->next;                //next entry
    }
    fprintf(fo, "; End of Address Table\n");
}

/* dump the decode (CFG) table */
void dumptab(void)
{
    DECODE *ptr;

    ptr = dec_head;                     //start at top
    printf("Decode table dump:\n");
    while (ptr) {
        printf("%c %04X %04X\n", ptr->type, ptr->start, ptr->end);
        ptr = ptr->next;                //next entry
    }
}

/* dump file in hexadecimal and ASCII */

void dumphex(void)
{
    int i, j;
    unsigned char byte;

    fprintf(fo, "; Hex/ASCII dump of file\n");
    fprintf(fo, "; Addr 0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F  0123456789ABCDEF\n");
    for (i=0; i<fsi; i+=16) {
	fprintf(fo, "; %04X ", i + sof);
	for (j=0; j<16; j++) {
            fseek(fi, i+j, SEEK_SET);
	    byte = fgetc(fi);
            if (i+j < fsi)
	        fprintf(fo, "%02X ", byte);
	}
	for (j=0; j<16; j++) {
            fseek(fi, i+j, SEEK_SET);
	    byte = fgetc(fi);
            if (i+j < fsi)
	        if (isprint(byte))
                    if (byte == '\t')
		        fprintf(fo, ".");
                    else
		        fprintf(fo, "%c", byte);
	        else
		    fprintf(fo, ".");
	}
	fprintf(fo, "\n");
    }
}

/* command -  process commands from cfg file */

void command()
{
    int i, lincnt = 1;
    char buf[128], cmd[10];
    unsigned int val;

    call_reg = -1; ret_reg = -1;

    while (fgets(buf, 127, fc)) {       //while more commands
        printf("line % 3d Cmd = %s", lincnt, buf);
        sscanf(buf, " %s %x", &cmd, &val);
        switch (toupper(cmd[0])) {
	    	case '#':
	    	case ';':
                break;
            case 'C':
                setcode('C', val);
                break;
            case 'D':
                setcode('D', val);
                break;
            case 'H':
                setcode('H', 0);
                dumphex();
                break;
            case 'L':
                if (val <=15) call_reg = val;
                break;
            case 'N':
                if (val <=15) ret_reg = val;
                break;
            case 'M':
                setcode('M', val);
                mem = val;
                break;
            case 'O':
                if (lincnt == 1) {
                    sof = val;          //set origin
                    eof += val;         //adjust eof
                    mem += val;         //adjust mem
		    printf("; Origin set to %04XH\n", sof);
		    fprintf(fo, "; Origin set to %04XH\n", sof);
		    fprintf(fo, "\tORG\t0%04XH\n", sof);
                } else
                    printf("Origin command must be first\n");
                break;
            case 'Q':
                return;
            case 'R':
                setcode('R', val);
                rom = val;
                break;
            case 'S':
                setcode('S', 0);
                break;
            case 'T':
                setcode('T', val);
                break;
            case 'X':
				setcode('X', val);
				break;
            default:
                printf("Bad command '%s' at line %d\n", cmd, lincnt);
                break;
        }
        lincnt++;
    }
}

/*  setcode - add table entry for the segment */

void setcode(unsigned char type, unsigned short addr)
{
    DECODE *nptr, *ptr;

    if ((nptr = (DECODE *) malloc(sizeof(DECODE))) == NULL) {
        printf("Unable to allocate memory for table entry\n");
        exit(1);
    }
    nptr->next = NULL;	                //initialize this entry
    nptr->type = type;
    nptr->start = addr;
    nptr->end = 0;
    if (dec_head == NULL)               //empty list, first entry
        dec_head = nptr;
    else {
        ptr = dec_head;                 //start at top
        while (ptr->next)               //find last entry
            ptr = ptr->next;
        ptr->next = nptr;               //link new one at bottom
        ptr->end = nptr->start;         //set ending address
    }
}

/* end of disasm80.c */
